from hitchstory.exceptions import Failure
from difflib import ndiff
import json


def _should_be_string(variable, type_of_string):
    if not isinstance(variable, str):
        raise Failure(
            (
                "strings_match(expected, actual) - {} should be a string. "
                "Instead it is of type {}.".format(
                    type_of_string,
                    str(type(variable)),
                )
            )
        )


def strings_match(expected, actual):
    """
    Check expected string matches actual, raise Failure if they do not.
    """
    _should_be_string(expected, "expected")
    _should_be_string(actual, "expected")

    if expected != actual:
        raise Failure(
            "ACTUAL:\n{0}\n\nEXPECTED:\n{1}\n\nDIFF:\n{2}".format(
                actual,
                expected,
                "".join(ndiff(expected.splitlines(1), actual.splitlines(1))),
            )
        )


def _json_should_be_string(variable, type_of_string):
    if not isinstance(variable, str):
        if isinstance(variable, dict):
            raise Failure(
                (
                    "json_match(expected, actual) - {} is a dict. It "
                    "should be a string of unparsed JSON.\n\n"
                    "You could try sending the original JSON string "
                    "or use json.dumps(the_dict)"
                ).format(type_of_string)
            )
        else:
            raise Failure(
                (
                    "json_match(expected, actual) - {} should be"
                    "a string, instead it is of type {}.".format(
                        type_of_string, str(type(variable))
                    )
                )
            )


def json_match(expected, actual):
    """
    Check expected JSON string matches actual, raise Failure if they do not.
    """
    _json_should_be_string(expected, "expected")
    _json_should_be_string(actual, "actual")

    try:
        expected_parsed_json = json.loads(expected)
    except json.decoder.JSONDecodeError:
        raise Failure(
            "json_match(expected, actual) - expected value is not valid JSON:\n\n{}".format(
                expected
            )
        )

    try:
        actual_parsed_json = json.loads(actual)
    except json.decoder.JSONDecodeError:
        raise Failure(
            "json_match(expected, actual) - actual value is not valid JSON:\n\n{}".format(
                actual
            )
        )

    expected_cleaned_json = json.dumps(expected_parsed_json, sort_keys=True, indent=4)

    actual_cleaned_json = json.dumps(actual_parsed_json, sort_keys=True, indent=4)

    if expected_parsed_json != actual_parsed_json:
        raise Failure(
            "ACTUAL:\n{0}\n\nEXPECTED:\n{1}\n\nDIFF:\n{2}".format(
                actual_cleaned_json,
                expected_cleaned_json,
                "".join(
                    ndiff(
                        expected_cleaned_json.splitlines(1),
                        actual_cleaned_json.splitlines(1),
                    )
                ),
            )
        )



from hitchstory.utils import DEFAULT_STACK_TRACE, underscore_slugify
from hitchstory.result import Success, Failure, FlakeResult
from hitchstory.story_step import StoryStep
from hitchstory.given import Given
from hitchstory import exceptions
from hitchstory import docstory
from hitchstory import utils
from collections import OrderedDict
from slugify import slugify
import mergedeep
import colorama
import time


__tracebackhide__ = True  # If used from within pytest


class StoryInfo:
    def __init__(self, info_definition, data):
        self._info = OrderedDict()
        for info_property in info_definition.keys():
            if info_property in data.keys():
                self._info[underscore_slugify(info_property)] = data.get(
                    underscore_slugify(info_property)
                )

    def get(self, key, default=None):
        return self._info.get(underscore_slugify(key), default)

    def items(self):
        return self._info.items()

    def __getitem__(self, key):
        return self._info[underscore_slugify(key)]

    def __contains__(self, key):
        return underscore_slugify(key) in self._info.keys()


class Story(object):
    def __init__(self, story_file, name, parsed_yaml, variation_of=None):
        self._story_file = story_file
        self._name = name
        self._slug = None
        self._parsed_yaml = parsed_yaml
        self._steps = []
        self._parent = None
        self._variation_of = variation_of
        self._info = StoryInfo(self.engine.info_definition, self.data)
        self._collection = self._story_file.collection
        self.children = []
        self.about = self.data.get("about", "")

    def _unparameterized_preconditions(self, child=False):
        precondition_dict = (
            self.parent._unparameterized_preconditions()
            if self.parent is not None
            else OrderedDict()
        )
        if child:
            precondition_dict = OrderedDict()
        for name, precondition in self.data.get("given", OrderedDict()).items():
            if self.engine.given_definition.should_override(name):
                if name not in precondition_dict:
                    precondition_dict[name] = precondition
                else:
                    merged = mergedeep.merge({}, precondition_dict[name], precondition)
                    precondition_dict[name] = merged

            else:
                precondition_dict[name] = precondition
        return precondition_dict

    @property
    def _yaml_steps(self):
        step_list = self._parent_steps

        if "following_steps" in self._parsed_yaml:
            step_list.extend(self._parsed_yaml.get("following_steps", []))
        elif "replacement_steps" in self._parsed_yaml:
            step_list = []
            step_list.extend(self._parsed_yaml.get("replacement_steps", []))
        else:
            yaml_child_steps = self._parsed_yaml.get("steps", [])
            if len(self._parent_steps) > 0 and len(yaml_child_steps) > 0:
                raise exceptions.AmbiguousSteps(
                    "Since '{}' has steps, '{}' must have either "
                    "'replacement steps' or 'following steps'".format(
                        self.parent.name,
                        self.name,
                    )
                )
            step_list.extend(yaml_child_steps)
        return step_list

    @property
    def _parent_steps(self):
        return self.parent._yaml_steps if self.parent is not None else []

    def _parameterized_preconditions(self, unparameterized_preconditions):
        all_preconditions = unparameterized_preconditions

        for name, precondition in all_preconditions.items():
            if utils.is_parameter(precondition):
                all_preconditions[name] = self.params[
                    utils.parameter_name(precondition)
                ]
            else:
                all_preconditions[name] = precondition
        return all_preconditions

    def initialize(self):
        """
        Revalidate steps and parameterize stories.
        These things can only be done after the story hierarchy is set
        """
        self._given = Given(
            preconditions=self._parameterized_preconditions(
                self._unparameterized_preconditions()
            ),
            story=self,
            child_preconditions=self._parameterized_preconditions(
                self._unparameterized_preconditions(child=True),
            ),
            document_templates=self.engine.given_definition.document_templates,
        )

        number_of_parent_steps = len(self._parent_steps)

        self._steps = [
            StoryStep(
                self, parsed_step, index, index - number_of_parent_steps, self.params
            )
            for index, parsed_step in enumerate(self._yaml_steps)
        ]

    def rewriter(self, given_or_step, args):
        return self._story_file.rewriter(self, given_or_step, args)

    def update(self, step, kwargs):
        self._story_file.update(self, step, kwargs)

    @property
    def variation(self):
        return self._variation_of is not None

    @property
    def data(self):
        return self._parsed_yaml.data

    @property
    def based_on(self):
        return self.data["based_on"] if "based_on" in self.data else self._variation_of

    @property
    def story_file(self):
        return self._story_file

    @property
    def engine(self):
        return self._story_file.engine

    @property
    def filename(self):
        return self._story_file.filename

    @property
    def info(self):
        return self._info

    @property
    def child_name(self):
        return self._name

    @property
    def name(self):
        return (
            self._name
            if not self.variation
            else "{0}/{1}".format(self._variation_of, self._name)
        )

    @property
    def slug(self):
        if self._slug is None:
            self._slug = slugify(self.name)  # relatively slow method
        return self._slug

    @property
    def parent(self):
        return self._parent

    @property
    def params(self):
        param_dict = self.parent.params if self.parent is not None else {}
        for name, param in self.data.get("with", {}).items():
            param_dict[name] = param
        for name, param in self._collection._params.items():
            param_dict[name] = param
        return param_dict

    @property
    def given(self):
        return self._given

    @property
    def steps(self):
        return self._steps

    def _run_special_method(self, method, exception_to_raise, result=None):
        try:
            if result is None:
                method()
            else:
                method(result)
        except Exception:
            stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

            raise exception_to_raise(stack_trace)

    def _run_on_success(self):
        try:
            self.engine.on_success()
        except Exception:
            self._run_special_method(
                self.engine.tear_down, exceptions.TearDownException
            )
            stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

            raise exceptions.OnSuccessException(stack_trace)

    def _run_on_failure(self, result):
        try:
            self.engine.on_failure(result)
        except Exception:
            self._run_special_method(
                self.engine.tear_down, exceptions.TearDownException
            )
            stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

            raise exceptions.OnFailureException(stack_trace)

    def _play_single_story(self):
        start_time = time.time()
        self._collection.log(
            "RUNNING {0} in {1} ... ".format(self.name, self._story_file.filename),
            newline=False,
        )
        passed = False
        caught_exception = None
        try:
            from signal import signal, SIGINT, SIGTERM, SIGHUP, SIGQUIT

            signal(SIGINT, self.engine.on_abort)
            signal(SIGTERM, self.engine.on_abort)
            signal(SIGHUP, self.engine.on_abort)
            signal(SIGQUIT, self.engine.on_abort)

            current_step = None
            self.engine._given = self.given
            self.engine.story = self
            self.engine.set_up()

            for step in self.steps:
                current_step = step
                self.engine.current_step = current_step
                run_step_method = step.method()
                run_step_method()

                if hasattr(self.engine, "_aborted") and self.engine._aborted:
                    self._collection.log("Aborted")
                    self._run_special_method(
                        self.engine.tear_down, exceptions.TearDownException
                    )
                    return
            passed = True
        except Exception as exception:
            caught_exception = exception
            if current_step is not None and current_step.expect_exception(
                self.engine, caught_exception
            ):
                failure_stack_trace = (
                    DEFAULT_STACK_TRACE.only_the_exception().current_stacktrace()
                )
            else:
                failure_stack_trace = DEFAULT_STACK_TRACE.current_stacktrace()

        if passed:
            self._run_on_success()
            result = Success(self, time.time() - start_time)
            self.story_file.rewrite()
            self._collection.log(
                "{green}SUCCESS{reset_all} in {duration:.1f} seconds.".format(
                    green=colorama.Fore.GREEN,
                    reset_all=colorama.Style.RESET_ALL,
                    duration=result.duration,
                )
            )
        else:
            result = Failure(
                self,
                time.time() - start_time,
                caught_exception,
                current_step,
                failure_stack_trace,
            )
            self._run_on_failure(result)
            self._collection.log(
                (
                    "{red}{bright}FAILED in {duration:.1f} seconds.{reset_all}"
                    "\n\n{blue}{story_snippet}{reset_all}\n{stacktrace}"
                ).format(
                    red=colorama.Fore.RED,
                    bright=colorama.Style.BRIGHT,
                    duration=result.duration,
                    blue=colorama.Fore.BLUE,
                    reset_all=colorama.Style.RESET_ALL,
                    story_snippet=result.story_failure_snippet,
                    stacktrace=result.stacktrace,
                )
            )

        self._run_special_method(self.engine.tear_down, exceptions.TearDownException)

        if not passed and self._collection._external_test_runner:
            raise exceptions.StoryFailure(self._collection._output_handle.getvalue())

        if not self._collection._external_test_runner:
            return result

    def documentation(self):
        """Generate textual documentation from story."""
        doc_templates = self._collection._doc_templates
        if doc_templates is None:
            raise exceptions.WithDocumentationMissing(
                "Documentation templates missing. Did you use .with_documentation?"
            )
        return docstory.story_template(self, doc_templates)

    def play(self):
        """
        Run a story from beginning to end, time it and return
        a Result object.
        """
        if self._collection._flakecheck_times is None:
            return self._play_single_story()
        else:
            start_time = time.time()
            result = FlakeResult()

            for i in range(self._collection._flakecheck_times):
                result.append(self._play_single_story())

            if result.is_flaky:
                self._collection.log(
                    "\nFLAKINESS DETECTED in {0:.1f} seconds, {1:.0f}% of stories failed.".format(
                        time.time() - start_time, result.percentage_failures
                    )
                )
                return result
            else:
                self._collection.log(
                    (
                        "\nNO FLAKINESS DETECTED in {duration:.1f} seconds "
                        "after running '{name}' story {count} times."
                    ).format(
                        duration=time.time() - start_time,
                        name=self.name,
                        count=result.total_results,
                    )
                )
                return result

    def __repr__(self):
        return "Story('{0}')".format(self._slug)



from strictyaml import Map, Str, Optional, load
from hitchstory.utils import to_underscore_style, current_stack_trace_data
from hitchstory.exceptions import DocumentationTemplateError
from hitchstory.engine import BaseEngine
import jinja2

EXCEPTION_TEMPLATE = """\
Exception in '{name}' template.

{exception_type}
{exception_message}
"""


class DocTemplate(object):
    def __init__(self, engine, doc_yaml_template, extra):
        self._doc_yaml_template = doc_yaml_template
        self.jenv = jinja2.Environment(
            undefined=jinja2.StrictUndefined, loader=jinja2.BaseLoader
        )
        if extra is not None:
            self.jenv.globals.update(extra)
        self._engine = engine

    def _step_methods(self):
        return [
            method.replace("_", " ")
            for method in self._engine.__class__.__dict__.keys()
            if not method.startswith("_") and method not in BaseEngine.__dict__.keys()
        ]

    def parse(self):
        self._parsed = load(
            self._doc_yaml_template,
            Map(
                {
                    "story": Str(),
                    Optional("variation"): Str(),
                    Optional("given"): Map(
                        {
                            Optional(name): Str()
                            for name in self._engine.given_definition.keys()
                        }
                    ),
                    Optional("info"): Map(
                        {
                            Optional(name): Str()
                            for name in self._engine.info_definition.keys()
                        }
                    ),
                    Optional("steps"): Map(
                        {Optional(name): Str() for name in self._step_methods()}
                    ),
                }
            ),
        ).data

        self._slugified_steps = {
            to_underscore_style(name): text
            for name, text in self._parsed.get("steps", {}).items()
        }

    def _render(self, name, template, variables):
        if template is not None:
            try:
                return self.jenv.from_string(template).render(**variables)
            except Exception:
                stack_trace = current_stack_trace_data()
                raise DocumentationTemplateError(
                    EXCEPTION_TEMPLATE.format(
                        name=name,
                        exception_type=stack_trace["exception_type"],
                        exception_message=stack_trace["exception_string"],
                    )
                )
        else:
            return ""

    def variation(self, **variables):
        return self._render(
            name="variation",
            template=self._parsed.get("variation"),
            variables=variables,
        )

    def story(self, **variables):
        return self._render(
            name="story", template=self._parsed["story"], variables=variables
        )

    def info_from_name(self, name, info_property):
        return self._render(
            name=f"info/{name}",
            template=self._parsed.get("info", {}).get(name),
            variables={name: info_property},
        )

    def given_from_name(self, name, given_property):
        return self._render(
            name=f"given/{name}",
            template=self._parsed.get("given", {}).get(name),
            variables={name: given_property},
        )

    def step_from_slug(self, slug, variables):
        return self._render(
            name=f"steps/{slug}",
            template=self._slugified_steps.get(slug),
            variables=variables,
        )



from strictyaml import load, Map, Str, Seq, Optional, MapPattern, Any, YAMLError
from hitchstory import exceptions, utils
from hitchstory.story import Story
from hitchstory.given import Given
from path import Path
import copy


class Rewriter(object):
    def __init__(self, story_file, story, given_or_step, args):
        self._story_file = story_file
        self._story = story
        self._given_or_step = given_or_step
        self._args = args

    @property
    def updated_yaml(self):
        if self._story_file._updated_yaml is None:
            self._story_file._updated_yaml = copy.deepcopy(
                self._story_file._parsed_yaml
            )
        return self._story_file._updated_yaml

    def to(self, text):
        key_to_update = self._args[-1]
        if self._story.variation:
            variations = self.updated_yaml[self._story.based_on]["variations"]

            if isinstance(self._given_or_step, Given):
                yaml_story = variations[self._story.child_name]
                given_prop = yaml_story["given"]

                for sub_prop in self._args[:-1]:
                    given_prop = given_prop[sub_prop]

                given_prop[self._args[-1]] = text
            else:
                if self._given_or_step.child_index >= 0:
                    yaml_story = variations[self._story.child_name]

                    step_type = _step_type(yaml_story)

                    if self._given_or_step.arguments.single_argument:
                        yaml_story[step_type][self._given_or_step.child_index][
                            self._given_or_step.name
                        ] = text
                    else:
                        yaml_step = yaml_story[step_type][
                            self._given_or_step.child_index
                        ][self._given_or_step.name]

                        for subkey in self._args[:-1]:
                            yaml_step = yaml_step[subkey]

                        yaml_step[key_to_update] = text
                else:
                    yaml_story = variations[self._story.child_name]
                    step_type = _step_type(yaml_story)

                    if self._given_or_step.arguments.single_argument:
                        yaml_story[step_type][self._given_or_step.index][
                            self._given_or_step.name
                        ] = text
                    else:
                        yaml_step = yaml_story[step_type][self._given_or_step.index][
                            self._given_or_step.name
                        ]

                        for subkey in self._args[:-1]:
                            yaml_step = yaml_step[subkey]

                        yaml_step[key_to_update] = text
        else:
            yaml_story = self.updated_yaml[self._story.name]

            if isinstance(self._given_or_step, Given):
                given_prop = yaml_story["given"]

                for sub_prop in self._args[:-1]:
                    given_prop = given_prop[sub_prop]

                given_prop[self._args[-1]] = text
            else:
                if "following_steps" in yaml_story:
                    step_to_update = yaml_story[_step_type(yaml_story)][
                        self._given_or_step.child_index
                    ]
                else:
                    step_to_update = yaml_story[_step_type(yaml_story)][
                        self._given_or_step.index
                    ]

                if self._given_or_step.arguments.single_argument:
                    single_argument_name = (
                        self._given_or_step.step_method_obj.single_argument_name
                    )
                    if key_to_update == single_argument_name:
                        step_to_update[self._given_or_step.name] = text
                    else:
                        raise exceptions.RewriteFailure(
                            (
                                f"'{key_to_update}' doesn't exist, "
                                f"only '{single_argument_name}' exists."
                            )
                        )
                else:
                    yaml_step = step_to_update[self._given_or_step.name]

                    for subkey in self._args[:-1]:
                        yaml_step = yaml_step[subkey]

                    yaml_step[key_to_update] = text


def _step_type(yaml_story):
    """Determine the step type of a story YAML snippet."""
    if "steps" in yaml_story:
        return "steps"
    elif "replacement_steps" in yaml_story:
        return "replacement_steps"
    elif "following_steps" in yaml_story:
        return "following_steps"
    else:
        raise NotImplementedError(
            "This should never happen - please raise "
            "a ticket on github.com/hitchdev/hitchstory"
        )


class StoryFile(object):
    """
    YAML file containing one or more named stories, part of a collection.
    """

    def __init__(self, filename, collection):
        self._filename = Path(filename).absolute()
        self._yaml = self._filename.bytes().decode("utf8")
        self._collection = collection
        self._updated_yaml = None

        steps_schema = Seq(Str() | MapPattern(Str(), Any(), maximum_keys=1))

        story_schema = {
            Optional("replacement_steps"): steps_schema,
            Optional("following_steps"): steps_schema,
            Optional("steps"): steps_schema,
            Optional("about"): Str(),
            Optional("with"): Any(),
            Optional("given"): self.engine.given_definition.preconditions,
        }

        # TODO : Make copy of story schema
        variation_schema = {
            Optional("replacement_steps"): steps_schema,
            Optional("following_steps"): steps_schema,
            Optional("steps"): steps_schema,
            Optional("about"): Str(),
            Optional("with"): Any(),
            Optional("given"): self.engine.given_definition.preconditions,
        }

        for info_property, info_property_schema in self.engine.info_definition.items():
            story_schema[Optional(info_property)] = info_property_schema
            variation_schema[Optional(info_property)] = info_property_schema

        story_schema[Optional("based_on")] = Str()
        story_schema[Optional("variations")] = MapPattern(
            Str(), Map(variation_schema, key_validator=utils.UnderscoredSlug())
        )

        try:
            self._parsed_yaml = load(
                self._yaml,
                Str()
                | MapPattern(
                    Str(), Map(story_schema, key_validator=utils.UnderscoredSlug())
                ),
            )
        except YAMLError as error:
            raise exceptions.StoryYAMLError(filename, str(error))

    @property
    def engine(self):
        return self._collection.engine

    def rewrite(self):
        """
        Rewrite all changes back to the file.
        """
        if self._updated_yaml is not None:
            self.path.write_text(self._updated_yaml.as_yaml())

    def rewriter(self, story, given_or_step, args):
        return Rewriter(
            self,
            story,
            given_or_step,
            args,
        )

    def update(self, story, step, kwargs):
        """
        Update a specific step in a particular story during a test run.
        """
        if self._updated_yaml is None:
            self._updated_yaml = copy.copy(self._parsed_yaml)

        if story.variation:
            if step.child_index >= 0:
                variations = self._updated_yaml[story.based_on]["variations"]
                yaml_story = variations[story.child_name]

                step_type = _step_type(yaml_story)

                if step.arguments.single_argument:
                    value = list(kwargs.values())[0]
                    yaml_story[step_type][step.child_index][step.name] = value
                else:
                    for key, value in kwargs.items():
                        yaml_story[step_type][step.child_index][step.name][key] = value
            else:
                yaml_story = self._updated_yaml[story.based_on]

                step_type = _step_type(yaml_story)

                if step.arguments.single_argument:
                    value = list(kwargs.values())[0]
                    yaml_story[step_type][step.index][step.name] = value
                else:
                    for key, value in kwargs.items():
                        yaml_story[step_type][step.index][step.name][key] = value
        else:
            yaml_story = self._updated_yaml[story.name]
            step_to_update = yaml_story[_step_type(yaml_story)][step.index]

            if step.arguments.single_argument:
                value = list(kwargs.values())[0]
                step_to_update[step.name] = value
            else:
                for key_to_update, value_to_update in kwargs.items():
                    step_to_update[step.name][key_to_update] = value_to_update

    @property
    def filename(self):
        return self._filename

    @property
    def collection(self):
        return self._collection

    @property
    def path(self):
        return Path(self._filename)

    def ordered_by_file(self):
        """
        Return all of the stories in the file in the order they
        appear in the file.
        """
        stories = []
        if self._parsed_yaml.is_mapping():
            for name, parsed_main_story in self._parsed_yaml.items():
                base_story = Story(self, str(name), parsed_main_story)
                stories.append(base_story)
                variations = []

                for variation_name, parsed_var_name in (
                    self._parsed_yaml[name].get("variations", {}).items()
                ):
                    variations.append(
                        Story(
                            self,
                            variation_name,
                            parsed_var_name,
                            variation_of=str(name),
                        )
                    )

                stories.extend(variations)
                base_story.variations = variations
        return stories



"""
User-exposed engine related code.
"""
from hitchstory import exceptions
from strictyaml import Any, Map, Optional
from hitchstory import utils


def validate(**kwargs):
    """
    Decorator for validating arguments in a HitchStory step.
    """

    def decorator(step_function):
        for arg in kwargs:
            if arg not in step_function.__code__.co_varnames:
                raise exceptions.StepContainsInvalidValidator(
                    "Step {} does not contain argument '{}' listed as a validator.".format(
                        step_function.__repr__(), arg
                    )
                )
        step_function._validators = kwargs
        return step_function

    return decorator


def no_stacktrace_for(exception_type):
    """
    Suppress stack trace for exceptions of exception_type on the
    method.
    """

    def decorator(step_function):
        if not hasattr(step_function, "_expected_exceptions"):
            step_function._expected_exceptions = [exception_type]
        else:
            step_function._expected_exceptions.append(exception_type)
        return step_function

    return decorator


def about(template):
    """
    Attach a template to a step which can be used to generate
    documentation about the step.
    """

    def decorator(step_function):
        step_function._about_template = template
        return step_function

    return decorator


class GivenProperty(object):
    OVERRIDE = -1
    REPLACE = -2

    def __init__(self, schema=None, document=None, inherit_via=None):
        self.document = document
        self.inherit_via = inherit_via

        if utils.is_schema_mapping(schema):
            if inherit_via == -1:
                self.schema = utils.optionalize_schema(schema)
            elif inherit_via == -2:
                self.schema = schema
            else:
                raise exceptions.InheritViaRequired(
                    "inherit_via is required on every GivenProperty that has a "
                    "strictyaml mapping schema (i.e. Map or MapPattern)."
                )
        else:
            if inherit_via is None:
                self.schema = Any() if schema is None else schema
            else:
                raise exceptions.InheritViaDisallowed(
                    "inherit_via cannot be used on non mapping-schemas "
                    "(i.e. every schema that isn't Map or MapPattern)."
                )


class GivenDefinition(object):
    def __init__(self, **given_properties):
        self.given_properties = given_properties
        mapping = {}
        for name, given_property in given_properties.items():
            assert isinstance(
                given_property, GivenProperty
            ), "{0} must be GivenProperty.".format(name)
            mapping[Optional(name)] = utils.YAML_Param | given_property.schema
        self.preconditions = Map(mapping, key_validator=utils.UnderscoredSlug())

    @property
    def document_templates(self):
        return {name: prop.document for name, prop in self.given_properties.items()}

    def keys(self):
        return self.given_properties.keys()

    def should_override(self, name):
        return self.given_properties[name].inherit_via == -1


class InfoProperty(object):
    def __init__(self, schema=None):
        self.schema = Any() if schema is None else schema


class InfoDefinition(object):
    def __init__(self, **info_properties):
        self._properties = {}
        for name, info_property in info_properties.items():
            assert isinstance(
                info_property, InfoProperty
            ), "{0} must be InfoProperty.".format(name)
            assert name not in [
                "based_on",
                "steps",
                "about",
                "with",
                "given",
                "variations",
                "examples",
            ], "{0} is not a valid name for an InfoProperty."
            self._properties[name] = info_property.schema

    def items(self):
        return self._properties.items()

    def keys(self):
        return self._properties.keys()


class NewStory(object):
    def __init__(self, engine):
        self._engine = engine

    def save(self):
        """
        Write out the updated story to file.
        """
        self._engine.story.story_file.rewrite()


class BaseEngine(object):
    given_definition = GivenDefinition()
    info_definition = InfoDefinition()

    @property
    def new_story(self):
        return NewStory(self)

    @property
    def given(self):
        return self._given

    def set_up(self):
        pass

    def on_abort(self, signal_num, stack_frame):
        self._aborted = True

    def on_failure(self, report):
        pass

    def on_success(self):
        pass

    def tear_down(self):
        pass



from strictyaml import Regex, ScalarValidator, validators
import strictyaml as sy
from re import compile
from path import Path
import prettystack
from jinja2.environment import Environment
from jinja2 import DictLoader
from slugify import slugify
import sys


SRC_DIR = Path(__file__).realpath().dirname()

TEMPLATE_DIR = SRC_DIR.joinpath("templates")


DEFAULT_STACK_TRACE = (
    prettystack.PrettyStackTemplate()
    .to_console()
    .cut_calling_code(SRC_DIR.joinpath("story.py"))
)


PARAM_REGEX = r"^\(\((.*?)\)\)$"


YAML_Param = Regex(PARAM_REGEX)


class UnderscoredSlug(ScalarValidator):
    def validate_scalar(self, chunk):
        return slugify(chunk.contents, separator="_")


def is_parameter(text):
    """
    Is the chunk of YAML data passed to us a parameter?

    i.e. like so (( parametername ))
    """
    return isinstance(text, str) and compile(PARAM_REGEX).match(text) is not None


def parameter_name(text):
    """
    Return parameter name from parameter text.

    e.g. (( param_name )) -> "param_name"
    """
    return compile(PARAM_REGEX).match(text).group(1).strip()


def to_underscore_style(text):
    """Changes "Something like this" to "something_like_this"."""
    text = text.lower().replace(" ", "_").replace("-", "_")
    return "".join(x for x in text if x.isalpha() or x.isdigit() or x == "_")


def underscore_slugify(text):
    """Changes "Something like this" to "something_like_this"."""
    return slugify(text, separator="_")


def render_template(templates_dict, template_name, parameters):
    """
    Render a jinja2 template.
    """
    env = Environment()
    env.loader = DictLoader(templates_dict)
    return env.get_template(template_name).render(**parameters)


def current_stack_trace_data():
    """
    Build a list of tracebacks from the last stack trace
    including line numbers and filenames.

    All the data needed to build a pretty stacktrace.
    """
    tb_id = 0
    _, exception, tb = sys.exc_info()

    if exception is None:
        return None

    # Create list of tracebacks
    tracebacks = []
    while tb is not None:
        filename = tb.tb_frame.f_code.co_filename
        if filename == "<frozen importlib._bootstrap>":
            break

        tracebacks.append(
            {
                "tb_id": tb_id,
                "filename": tb.tb_frame.f_code.co_filename,
                "line": tb.tb_lineno,
                "function": tb.tb_frame.f_code.co_name,
            }
        )

        tb_id = tb_id + 1
        tb = tb.tb_next

    return {
        "tracebacks": tracebacks,
        "exception_string": str(exception),
        "docstring": exception.__doc__ if exception.__doc__ is not None else None,
        "exception_type": "{}.{}".format(
            type(exception).__module__, type(exception).__name__
        ),
    }


def is_schema_mapping(schema):
    return isinstance(schema, validators.MapValidator)


def optionalize_schema(schema):
    if isinstance(schema, sy.Map):
        return sy.Map(
            {
                key
                if isinstance(key, sy.Optional)
                else sy.Optional(key): optionalize_schema(value)
                for key, value in schema._validator.items()
            },
            key_validator=schema._key_validator,
        )
    else:
        return schema



from hitchstory import utils
from collections import OrderedDict
import jinja2


class GivenProperty(object):
    def __init__(self, value, documentation):
        self._value = value
        self._documentation = documentation

    @property
    def value(self):
        return self._value

    @property
    def documentation(self):
        return self._documentation


class GivenProperties(object):
    def __init__(self, given):
        self._given = given

        self._properties = OrderedDict()
        for name, precondition in self._given.items():
            self._properties[name] = GivenProperty(
                precondition,
                self._given.story,
                jinja2.Template(self._given._document_templates[name]).render(
                    **{name: precondition}
                ),
            )

    def items(self):
        return self._properties.items()


class Given(object):
    def __init__(
        self, preconditions, story, child_preconditions=None, document_templates=None
    ):
        self._preconditions = preconditions
        self.story = story

        if child_preconditions is not None:
            self.child = Given(child_preconditions, story)
        else:
            self.child = self

        self._document_templates = document_templates

    def rewrite(self, *args):
        return self.story.rewriter(self, args)

    def get(self, key, default=None):
        return self._preconditions.get(utils.underscore_slugify(key), default)

    def __getitem__(self, key):
        slug = utils.underscore_slugify(key)

        if slug in self._preconditions:
            return self._preconditions[slug]
        else:
            raise KeyError(
                (
                    "'{}' / '{}' not found from given. Preconditions available: {}"
                ).format(
                    key,
                    slug,
                    ", ".join(self._preconditions.keys())
                    if len(self._preconditions.keys()) > 0
                    else "None",
                )
            )

    def __contains__(self, key):
        return utils.underscore_slugify(key) in self._preconditions.keys()

    def keys(self):
        return self._preconditions.keys()

    def items(self):
        return self._preconditions.items()

    def values(self):
        return self._preconditions.values()

    @property
    def properties(self):
        return GivenProperties(self)



from hitchstory.story_list import StoryList
from hitchstory.story_file import StoryFile
from hitchstory.doc_template import DocTemplate
from hitchstory.engine import BaseEngine
from collections import OrderedDict
from hitchstory import exceptions
from slugify import slugify
from path import Path
from copy import copy
import io
import sys


class StoryCollection(object):
    """
    Unordered group of related stories.
    """

    def __init__(self, storypaths, engine):
        """
        Create a collection of hitch stories from a list (or generator) of
        story paths and an initialized story engine.
        """
        if not isinstance(engine, BaseEngine):
            raise exceptions.WrongEngineType(
                "Engine should inherit from hitchstory.BaseEngine."
            )
        if isinstance(storypaths, str):
            raise exceptions.InvalidStoryPaths(
                (
                    "storypaths should be a list or iterator returning a list of story files"
                    " (e.g. using pathlib.Path.glob). Instead it was string '{0}'."
                ).format(storypaths)
            )
        self._storypaths = storypaths
        self._engine = engine
        self._in_filename = None
        self._named = None
        self._filters = []
        self._params = {}
        self._story_files = {}
        self._filtered_stories = None
        self._non_variations = False
        self._only_uninherited = False
        self._templates = {}
        self._output_handle = sys.stdout
        self._flakecheck_times = None
        self._doc_templates = None
        self._external_test_runner = False

    @property
    def engine(self):
        return self._engine

    def story_file(self, filename):
        return StoryFile(filename, self)

    @property
    def stories(self):
        story_dict = OrderedDict()
        for filename in self._storypaths:
            if not Path(filename).exists():
                raise exceptions.InvalidStoryPaths(
                    "Story file name '{0}' does not exist.".format(filename)
                )
            if Path(filename).is_dir():
                raise exceptions.InvalidStoryPaths(
                    "Story path '{0}' is a directory.".format(filename)
                )
            for story in self.story_file(filename).ordered_by_file():
                if story.slug in story_dict:
                    raise exceptions.DuplicateStoryNames(story, story_dict[story.slug])
                story_dict[story.slug] = story

        # Make sure parent stories know who their children are and vice versa
        for name, story in story_dict.items():
            if story.based_on is not None:
                parent_slug = slugify(story.based_on)

                if parent_slug not in story_dict:
                    raise exceptions.BasedOnStoryNotFound(
                        story.based_on, story.name, story.filename
                    )
                story_dict[parent_slug].children.append(story)
                story._parent = story_dict[parent_slug]

        # Revalidate steps and parameterize
        # These things can only be done after the story hierarchy is set
        for story in story_dict.values():
            story.initialize()
        return story_dict

    def ordered_by_file(self):
        """
        Return a StoryList containing stories filtered from the
        collection.

        The stories are primarily ordered according to the file
        ordering given to the story collection and secondarily
        according to the order they appear in those files.
        """
        return StoryList(self.ordered_arbitrarily())

    def ordered_arbitrarily(self):
        """
        Return a StoryList object containing stories filtered
        from the collection.
        """
        if self._filtered_stories is None:
            self._filtered_stories = []
            all_stories = []
            for story in self.stories.values():
                filtered = True
                for filter_func in self._filters:
                    if not filter_func(story):
                        filtered = False
                if self._named is not None:
                    if story.name != self._named:
                        filtered = False
                if self._in_filename is not None:
                    if (
                        Path(story.filename).absolute()
                        != Path(self._in_filename).absolute()
                    ):
                        filtered = False
                if self._non_variations:
                    if story.variation:
                        filtered = False
                if self._only_uninherited:
                    if len(story.children) > 0:
                        filtered = False
                if filtered:
                    self._filtered_stories.append(story)
                all_stories.append(story)
        return self._filtered_stories

    def filter(self, filter_func):
        assert callable(filter_func)
        new_collection = self.copy()
        new_collection._filters.append(filter_func)
        return new_collection

    def in_filename(self, filename):
        new_collection = self.copy()
        new_collection._in_filename = Path(filename)
        if not new_collection._in_filename.exists():
            raise exceptions.FileNotFound(filename)
        return new_collection

    def with_params(self, **params):
        new_collection = self.copy()
        new_collection._params = params
        return new_collection

    def without_filters(self):
        new_collection = self.copy()
        new_collection._filters = []
        return new_collection

    def in_any_filename(self):
        new_collection = self.copy()
        new_collection._in_filename = None
        return new_collection

    def non_variations(self):
        new_collection = self.copy()
        new_collection._non_variations = True
        return new_collection

    def only_uninherited(self):
        new_collection = self.copy()
        new_collection._only_uninherited = True
        return new_collection

    def with_flake_detection(self, times=5):
        new_collection = self.copy()
        new_collection._flakecheck_times = times
        return new_collection

    def with_templates(self, templates):
        new_collection = self.copy()
        new_collection._templates = templates
        return new_collection

    def with_documentation(self, yaml_documentation, extra=None):
        new_collection = self.copy()
        doc_template = DocTemplate(self.engine, yaml_documentation, extra)
        doc_template.parse()
        new_collection._doc_templates = doc_template
        return new_collection

    def with_external_test_runner(self):
        new_collection = self.copy()
        new_collection._output_handle = io.StringIO()
        new_collection._external_test_runner = True
        return new_collection

    def copy(self):
        new_collection = copy(self)
        new_collection._filtered_stories = None
        new_collection._stories = None

        for story in new_collection._story_files.values():
            story._collection = self
        return new_collection

    def named(self, name):
        """
        Return a single story object with the name specified.

        Only slugified names are compared. E.g. "Story NAME" and "story-name" are equivalent.
        """
        slug = slugify(name)
        for story in self.ordered_arbitrarily():
            if slug == story.slug:
                return story
        raise exceptions.StoryNotFound(name)

    def ordered_by_name(self):
        """
        Return a list of stories ordered by name.
        """
        return StoryList(
            sorted(self.ordered_arbitrarily(), key=lambda story: story.name)
        )

    def shortcut(self, *words):
        """
        Return a single story that matches all of the words.
        """
        matching = []
        stories = self.ordered_by_name()
        slugified_words = [slugify(word) for word in words]
        for story in stories:
            if len(words) == len(
                [
                    slugified_word
                    for slugified_word in slugified_words
                    if slugified_word in story.slug
                ]
            ):
                matching.append(story)
        if len(matching) == 0:
            raise exceptions.StoryNotFound(", ".join(words))
        elif len(matching) > 1:
            raise exceptions.MoreThanOneStory(
                "\n".join(
                    [
                        "{0} (in {1})".format(story.name, story.filename)
                        for story in matching
                    ]
                )
            )
        else:
            return matching[0]

    def log(self, message, newline=True):
        """
        Log message to the output handle (usually stdout).
        """
        self._output_handle.write("{0}{1}".format(message, "\n" if newline else ""))
        self._output_handle.flush()

    def one(self):
        stories = self.ordered_by_name()
        if len(stories) > 1:
            raise exceptions.MoreThanOneStory(
                "\n".join(
                    [
                        "{0} (in {1})".format(story.name, story.filename)
                        for story in stories
                    ]
                )
            )
        elif len(stories) == 0:
            raise exceptions.NoStories()
        else:
            return stories[0]



"""Documentation objects for use in templates."""
from hitchstory.step_method import StepMethod
import pathlib


class DocStory(object):
    def __init__(self, story):
        self.slug = story.slug
        self.name = story.name


class DocVariation(object):
    def __init__(self, templates, variables):
        self._documentation = templates.variation(**variables)

    def documentation(self):
        return self._documentation


class DocInfoProperty(object):
    def __init__(self, templates, name, info_property):
        self._documentation = templates.info_from_name(name, info_property)

    def documentation(self):
        return self._documentation


class DocGivenProperty(object):
    def __init__(self, templates, name, given_property):
        self._documentation = templates.given_from_name(name, given_property)

    def documentation(self):
        return self._documentation


class DocGivenProperties(object):
    def __init__(self, templates, given, child=False):
        self._property_docs = {
            name: DocGivenProperty(templates, name, given[name])
            for name, given_property in given.items()
        }
        if not child:
            self.child = DocGivenProperties(templates, given.child, child=True)

    def __getattr__(self, name):
        return self._property_docs[name]

    def items(self):
        return self._property_docs.items()


class DocStep(object):
    def __init__(self, templates, step, this_story):
        step_method = StepMethod(step.step_method)
        arguments = {name: None for name in step_method.argspec.args[1:]}

        if step.arguments.is_none:
            arguments = {}
        elif step.arguments.single_argument:
            var_name = step_method.argspec.args[1:][0]
            arguments[var_name] = step.arguments.data
        else:
            if step_method.argspec.keywords:
                var_name = step_method.argspec.keywords
                arguments[var_name] = step.arguments.data
            else:
                arguments.update(step.arguments.data)

        variables = dict(arguments)
        variables["this_step"] = self
        variables["this_story"] = this_story
        self._slug = step.slug
        self._index = step.index
        self._documentation = templates.step_from_slug(step.slug, variables)

    @property
    def slug(self):
        return self._slug

    @property
    def index(self):
        return self._index

    def documentation(self):
        return self._documentation


def _story_variables(story, doc_templates, variation=False):
    this_story = DocStory(story)
    variables = {
        "info": {
            name: DocInfoProperty(doc_templates, name, info_property)
            for name, info_property in story.info.items()
        },
        "slug": story.slug,
        "given": DocGivenProperties(doc_templates, story.given),
        "name": story.child_name if variation else story.name,
        "about": story.about,
        "steps": [DocStep(doc_templates, step, this_story) for step in story.steps],
        "filename": pathlib.Path(story.filename),
    }
    if variation:
        variables["full_name"] = story.name
    return variables


def story_template(story, doc_templates):
    variables = _story_variables(story, doc_templates)
    variables["variations"] = [
        DocVariation(
            doc_templates, _story_variables(variation, doc_templates, variation=True)
        )
        for variation in story.variations
    ]
    return doc_templates.story(**variables)
